theory LTS
imports 
  "$AFP/Abstract-Rewriting/Abstract_Rewriting"
  "$AFP/Show/Show_Instances"
  "$AFP/Derive/Comparator_Generator/Compare_Generator"
  Formula
begin

datatype 'v trans_var = Pre 'v | Post 'v | Intermediate 'v
derive "show" trans_var
derive compare_order trans_var

type_synonym ('f,'v,'t) transition_formula = "('f,'v trans_var,'t) formula"

datatype ('f,'v,'t,'l) transition =
 Transition (trans_source: 'l) (trans_formula: "('f,'v trans_var,'t) formula") (trans_target: 'l)

datatype ('v,'t,'d,'l) state = State "('v,'t,'d) valuation" 'l

fun relabel_state where "relabel_state f (State \<alpha> l) = State \<alpha> (f l)"

datatype ('f,'v,'t,'l,'tr) lts =
 Lts (initial: 'l) (transitions: "'tr set") (transition_of: "'tr \<Rightarrow> ('f,'v,'t,'l) transition")

datatype ('f,'v,'t,'l,'tr) safety_prog = Safety_Prog 'l "('f,'v,'t,'l,'tr) lts"

context pretheory begin

definition wt_transition :: "('f,'v,'t,'l) transition \<Rightarrow> bool" where
  "wt_transition \<tau> \<equiv> formula (trans_formula \<tau>)"

definition lts :: "('f,'v,'t,'l,'tr) lts \<Rightarrow> bool" where
  "lts P \<equiv> \<forall> \<tau> \<in> transitions P. wt_transition (transition_of P \<tau>)"

fun safety_prog :: "('f,'v,'t,'l,'tr) safety_prog \<Rightarrow> bool" where
  "safety_prog (Safety_Prog _ lp) = lts lp"

fun pre_post_inter_valuation (* :: 
  "'v valuation \<Rightarrow> 'v valuation \<Rightarrow> 'v valuation \<Rightarrow> 'v trans_var valuation"*) where
  "pre_post_inter_valuation pre post inte (Pre x,t) = pre (x,t)"
| "pre_post_inter_valuation pre post inte (Post x,t) = post (x,t)"
| "pre_post_inter_valuation pre post inte (Intermediate x,t) = inte (x,t)"

definition transition :: "('f,'v,'t,'l,'tr) lts \<Rightarrow> ('v,'t,'d,'l) state rel" where
  "transition P \<equiv> { (State \<alpha>1 l1, State \<alpha>2 l2) | l1 l2 \<alpha>1 \<alpha>2 \<phi> \<beta> \<tau>. 
    assignment \<alpha>1 \<and> assignment \<alpha>2 \<and> assignment \<beta> \<and>
    \<tau> \<in> transitions P \<and> transition_of P \<tau> = Transition l1 \<phi> l2 
    \<and> satisfies (pre_post_inter_valuation \<alpha>1 \<alpha>2 \<beta>) \<phi>}"

definition initial_states :: "('f,'v,'t,'l,'tr) lts \<Rightarrow> ('v,'t,'d,'l) state set" where
  "initial_states P = { State \<alpha> l | \<alpha> l. l = initial P \<and> assignment \<alpha>}"

definition lts_termination :: "('f,'v,'t,'l,'tr) lts \<Rightarrow> bool" where
  "lts_termination P \<equiv> SN_on (transition P) (initial_states P)"

definition reachable_states :: "('f,'v,'t,'l,'tr) lts \<Rightarrow> ('v,'t,'d,'l) state set" where
  "reachable_states P = (transition P)^* `` (initial_states P)"

definition safe :: "('f,'v,'t,'l,'tr) safety_prog \<Rightarrow> bool" where
  "safe SP \<equiv> case SP of Safety_Prog err P \<Rightarrow> (\<forall> \<alpha>. assignment \<alpha> \<longrightarrow> (State \<alpha> err) \<notin> reachable_states P)"

end

subsection {* An Implementation of Linear Programs *}

type_synonym ('f,'v,'t,'l,'tr) transitions_impl = "('tr \<times> ('f,'v,'t,'l) transition) list"

datatype ('f,'v,'t,'l,'tr) lts_impl = Lts_Impl 'l "('f,'v,'t,'l,'tr) transitions_impl"

fun lts_of :: "('f,'v,'t,'l,'tr) lts_impl \<Rightarrow> ('f,'v,'t,'l,'tr) lts" where
  "lts_of (Lts_Impl i ts) = (Lts i (set (map fst ts)) (the o map_of ts))"

(* TODO: improve efficiency! *)
fun succ_transitions :: "('f,'v,'t,'l,'tr) lts_impl \<Rightarrow> ('f,'v,'t,'l,'tr) lts \<Rightarrow> 'l \<Rightarrow> ('tr \<times> 'l)list" where
  "succ_transitions (Lts_Impl i ts) (Lts _ _ \<tau>s) l = 
    [ (\<tau>,l') . (\<tau>,_) \<leftarrow> ts, Transition ll \<phi> l' \<leftarrow> [\<tau>s \<tau>], ll = l]"

lemma succ_transitions: "set (succ_transitions Pi (lts_of Pi) l) = 
  {(\<tau>, trans_target \<tau>') |\<tau> \<tau>'.
    \<tau> \<in> transitions (lts_of Pi) \<and> \<tau>' = transition_of (lts_of Pi) \<tau> \<and> trans_source \<tau>' = l}"
  by (cases Pi, auto split: transition.splits; force?, (metis transition.sel transition.collapse)?)

context pretheory begin

(* TODO: a nicer checking function *)
definition "lts_impl Pi \<equiv> lts (lts_of Pi)"

definition "lts_impl_checker Pi = check (lts_impl Pi) (shows ''ill-formed transition system'')"

lemma lts_impl_checker[intro]: "isOK (lts_impl_checker Pi) \<Longrightarrow> lts_impl Pi"
  by(simp add: lts_impl_checker_def)

lemma lts_impl[intro]: "lts_impl Pi \<Longrightarrow> lts (lts_of Pi)" 
  unfolding lts_impl_def by auto

end

section{* Corporation Graphs *}


datatype 'l copy = Copy 'l | Cutpoint_Aux 'l | Error_Loc

datatype ('f,'v,'t,'l,'tr) cooperation_prog = Cooperation_Prog
  (orig_prog: "('f,'v,'t,'l,'tr) lts")
  (cutpoints: "'l set")
  (deleted_transitions: "('f,'v,'t,'l copy) transition set")

fun copy_prog :: "('f,'v,'t,'l,'tr) cooperation_prog \<Rightarrow> ('f,'v,'t,'l copy,'tr) lts" where
  "copy_prog P = undefined"

fun complete_copy_prog :: "('f,'v,'t,'l,'tr) cooperation_prog \<Rightarrow> ('f,'v,'t,'l + 'l copy,'tr) lts" where
  "complete_copy_prog P = undefined"

context pretheory begin

definition cooperation_prog_termination :: "('f,'v,'t,'l,'tr) cooperation_prog \<Rightarrow> bool" where
  "cooperation_prog_termination P \<equiv> SN_on (transition (copy_prog P))
    (relabel_state Copy ` reachable_states (orig_prog P))"

definition step1 :: "('f,'v,'t,'l,'tr) lts \<Rightarrow> ('f,'v,'t,'l,'tr) cooperation_prog option" where
  "step1 = undefined"

lemma step1: "step1 P = Some Q \<Longrightarrow> cooperation_prog_termination Q \<Longrightarrow> lin_prog_termination P"
  oops

definition step2 :: "('f,'v,'t,'l,'tr) cooperation_prog \<Rightarrow> ('f,'v,'t,'l,'tr) cooperation_prog option" where
  "step2 = undefined"

lemma step2: "step2 P = Some Q \<Longrightarrow> cooperation_prog_termination Q \<Longrightarrow> cooperation_prog_termination P"
  oops

type_synonym rank_function_info = string

definition step3 :: "rank_function_info \<Rightarrow> ('f,'v,'t,'l,'tr) cooperation_prog \<Rightarrow> ('f,'v2,'t,'l2,'tr) safety_prog option" where
  "step3 = undefined"

lemma step3: "step3 rfi P = Some SP \<Longrightarrow> safety SP \<Longrightarrow> cooperation_prog_termination P"
  oops

end

end