theory Term_Autoref
imports
  "$AFP/Automatic_Refinement/Automatic_Refinement"
  "../Rewriting/Term"
begin
section {* Autoref setup for terms and related types *}

subsection {* Relator for @{typ "('f,'v)term"} *}

inductive_set term_rel_aux for Rs Rv where
  Var: "(x,x') \<in> Rv  \<Longrightarrow> (Var x, Var x') \<in> term_rel_aux Rs Rv"
| Fun: "\<lbrakk> (f,f') \<in> Rs;  list_all2 (\<lambda>t t'. (t, t') \<in> term_rel_aux Rs Rv) ts ts' \<rbrakk>
          \<Longrightarrow> (Fun f ts, Fun f' ts') \<in> term_rel_aux Rs Rv"

definition term_rel_def_internal: "term_rel \<equiv> term_rel_aux"
lemma term_rel_def[refine_rel_defs]: 
  "\<langle>Rs, Rv\<rangle>term_rel \<equiv> term_rel_aux Rs Rv"
by (simp add: term_rel_def_internal relAPP_def)

lemma term_rel_unfold:
  "list_all2 (\<lambda>t t'. (t, t') \<in> term_rel_aux F V) t t' \<longleftrightarrow> (t,t') \<in> \<langle>\<langle>F,V\<rangle>term_rel\<rangle>list_rel"
unfolding term_rel_def list_rel_def by blast

lemma term_rel_unfold':
  "list_all2 (\<lambda>t t'. (t, t') \<in> term_rel_aux F V \<and> P t t') t t' \<longleftrightarrow> (t,t') \<in> \<langle>\<langle>F,V\<rangle>term_rel\<rangle>list_rel \<and> list_all2 P t t'"
unfolding term_rel_def list_rel_def list_all2_conv_all_nth by auto

lemmas term_rel_Var[intro] = Var[folded term_rel_def]
lemmas term_rel_Fun[intro] = Fun[unfolded term_rel_unfold, folded term_rel_def]
lemmas term_rel_simps = term_rel_aux.simps[unfolded term_rel_unfold, folded term_rel_def]
lemmas term_rel_cases = term_rel_aux.cases[unfolded term_rel_unfold, folded term_rel_def]
lemmas term_rel_induct = term_rel_aux.induct[unfolded term_rel_unfold', folded term_rel_def]

lemma term_rel_aux_simps[simp]:
  "(Var x, Fun f ts) \<notin> \<langle>F,V\<rangle>term_rel"
  "(Fun f ts, Var x) \<notin> \<langle>F,V\<rangle>term_rel"
by (auto simp: term_rel_simps)

lemma term_relE1:
  assumes "(Var x, t) \<in> \<langle>F,V\<rangle>term_rel"
  obtains x' where "t = Var x'" "(x,x') \<in> V"
using assms by (auto simp: term_rel_simps)

lemma term_relE2:
  assumes "(t, Var x) \<in> \<langle>F,V\<rangle>term_rel"
  obtains x' where "t = Var x'" "(x',x) \<in> V"
using assms by (auto simp: term_rel_simps)

lemma term_relE3:
  assumes "(t, Fun f ts) \<in> \<langle>F,V\<rangle>term_rel"
  obtains f' ts' where "t = Fun f' ts'" "(f',f) \<in> F" "(ts',ts) \<in> \<langle>\<langle>F,V\<rangle>term_rel\<rangle>list_rel"
using assms by (auto simp: term_rel_simps)

lemma term_relE4:
  assumes "(Fun f ts, t) \<in> \<langle>F,V\<rangle>term_rel"
  obtains f' ts' where "t = Fun f' ts'" "(f,f') \<in> F" "(ts,ts') \<in> \<langle>\<langle>F,V\<rangle>term_rel\<rangle>list_rel"
using assms by (auto simp: term_rel_simps)

lemmas term_relE = term_relE1 term_relE2 term_relE3 term_relE4

lemma term_rel_sv[relator_props]:
  assumes "single_valued Rs" "single_valued Rv"
  shows "single_valued (\<langle>Rs, Rv\<rangle>term_rel)"
proof (intro single_valuedI allI impI)
  fix x y z
  assume *: "(x, y) \<in> \<langle>Rs, Rv\<rangle>term_rel" and "(x, z) \<in> \<langle>Rs, Rv\<rangle>term_rel"
  thus "y = z"
  proof (induction arbitrary: z rule: term_rel_induct[OF *])
    case (1 x x')
      with assms[THEN single_valuedD] show ?case by (auto simp: term_rel_simps)
    next
    case (2 f f' ts ts')
      note Fun = this
      from Fun(4) obtain fz tz where
        z: "z = Fun fz tz" and tz: "(ts,tz) \<in> \<langle>\<langle>Rs,Rv\<rangle>term_rel\<rangle>list_rel" by (auto simp: term_rel_simps)
      from Fun(2) tz have "ts' = tz" by (simp add: nth_equalityI list_rel_def list_all2_conv_all_nth z) 
      with Fun assms[THEN single_valuedD] show ?case by (auto simp: term_rel_simps z)
  qed
qed

lemma term_rel_id[relator_props]:
  assumes "Rs = Id" and "Rv = Id"
    shows "\<langle>Rs, Rv\<rangle>term_rel = Id"
proof(simp only: assms, rule)
  show "\<langle>Id, Id\<rangle>term_rel \<subseteq> Id"
  proof
    fix x y :: "('f, 'v) term"
    assume *: "(x,y) \<in> \<langle>Id, Id\<rangle>term_rel"
    have "x = y" by (induction rule: term_rel_induct[OF *], simp_all add: list_all2_eq)
    thus "(x,y) \<in> Id" by blast
  qed
  show "Id \<subseteq> \<langle>Id, Id\<rangle>term_rel"
  proof
    fix x y :: "('f, 'v) term"
    assume "(x,y) \<in> Id"
    moreover have "(x,x) \<in> \<langle>Id, Id\<rangle>term_rel" by (induction x, auto simp: list_rel_def list_all2_same)
    ultimately show "(x,y) \<in> \<langle>Id, Id\<rangle>term_rel" by simp
  qed
qed

lemma term_rel_id_simp[simp]:
  shows "\<langle>Id, Id\<rangle>term_rel = Id" by tagged_solver

lemma term_rel_mono[relator_props]:
  assumes "Rs \<subseteq> Rs'" and "Rv \<subseteq> Rv'"
  shows "\<langle>Rs, Rv\<rangle>term_rel \<subseteq> \<langle>Rs', Rv'\<rangle>term_rel"
proof
  fix x y
  assume *: "(x,y) \<in> \<langle>Rs, Rv\<rangle>term_rel"
  from assms show "(x,y) \<in> \<langle>Rs', Rv'\<rangle>term_rel"
  by (induction rule: term_rel_induct[OF *], auto simp: list_rel_def list_all2_conv_all_nth)
qed

lemma term_rel_range[relator_props]:
  assumes sym: "Range Rs = UNIV" and vars: "Range Rv = UNIV"
    shows "Range (\<langle>Rs, Rv\<rangle>term_rel) = UNIV"
proof
  show "Range (\<langle>Rs, Rv\<rangle>term_rel) \<subseteq> UNIV" by blast
  show "UNIV \<subseteq> Range (\<langle>Rs, Rv\<rangle>term_rel)"
  proof
    fix x
    from assms show "x \<in> Range (\<langle>Rs, Rv\<rangle>term_rel)" unfolding Range_iff proof (induction x)
      case (Fun f ts)
        hence "\<forall>t . \<exists>t'. t \<in> set ts \<longrightarrow> (t', t) \<in> \<langle>Rs, Rv\<rangle>term_rel" by blast
        from choice[OF this] obtain ts' where ts': "\<And>x. x \<in> set ts \<Longrightarrow> (ts' x, x) \<in> \<langle>Rs, Rv\<rangle>term_rel" by blast
        moreover from sym obtain f' where "(f',f) \<in> Rs" by auto
        ultimately show ?case by (intro exI[of _ "Fun f' (map ts' ts)"], auto simp: list_rel_def list_all2_conv_all_nth)
    qed (auto simp add: term_rel_simps)
  qed
qed

lemma term_autoref[autoref_rules, param]:
  "(Var, Var) \<in> V \<rightarrow> \<langle>F,V\<rangle>term_rel"
  "(Fun, Fun) \<in> F \<rightarrow> \<langle>\<langle>F,V\<rangle>term_rel\<rangle>list_rel \<rightarrow> \<langle>F,V\<rangle>term_rel"
  "(case_term, case_term) \<in> (V \<rightarrow> X) \<rightarrow> (F \<rightarrow> \<langle>\<langle>F,V\<rangle>term_rel\<rangle>list_rel \<rightarrow> X) \<rightarrow> \<langle>F,V\<rangle>term_rel \<rightarrow> X"
by (force split: term.split simp: term_rel_simps dest: fun_relD)+

lemma rec_term_autoref[autoref_rules, param]:
  "(rec_term, rec_term) \<in> (V \<rightarrow> X) \<rightarrow> (F \<rightarrow> \<langle>\<langle>F,V\<rangle>term_rel \<times>\<^sub>r X\<rangle>list_rel \<rightarrow> X) \<rightarrow> \<langle>F,V\<rangle>term_rel \<rightarrow> X"
proof (intro fun_relI, goal_cases)
case (1 v v' l l' t t')
  thus ?case proof (induction t arbitrary: t')
    case (Var x)
      from this obtain x' where t': "t' = Var x'" "(x,x') \<in> V" by (auto simp: term_rel_simps)
      with Var show ?case unfolding t' by (auto dest: fun_relD)
    next
    case (Fun f ts)
      let ?r = "rec_term v l" and ?r' = "rec_term v' l'"
      from Fun obtain f' ts' where t': "t' = Fun f' ts'" and f: "(f,f') \<in> F"
        and ts: "(ts, ts') \<in> \<langle>\<langle>F,V\<rangle>term_rel\<rangle>list_rel" by (auto simp: term_rel_simps)
      hence "length ts = length ts'" by (auto simp: list_rel_def dest: list_all2_lengthD)
      hence all: "list_all2 (\<lambda>t t'. (?r t, ?r' t') \<in> X) ts ts'"
        using Fun ts by (auto simp: list_rel_def list_all2_conv_all_nth)
      hence "(map (\<lambda>t. (t, ?r t)) ts, map (\<lambda>t'. (t', ?r' t')) ts') \<in> \<langle>\<langle>F, V\<rangle>term_rel \<times>\<^sub>r X\<rangle>list_rel"
        using ts by (induction rule: list_all2_induct, simp_all)
      thus ?case by (simp add: t', rule fun_relD[OF Fun.prems(2) f, THEN fun_relD])
  qed
qed

lemma map_rec_term: "map_term ff fv t = rec_term (Var o fv) (\<lambda>f ts. Fun (ff f) (map snd ts)) t"
by (induction t, auto)

lemma map_term_autoref[autoref_rules, param]:
  "(map_term, map_term) \<in> (F \<rightarrow> F') \<rightarrow> (V \<rightarrow> V') \<rightarrow> \<langle>F,V\<rangle>term_rel \<rightarrow> \<langle>F',V'\<rangle>term_rel"
by (intro fun_relI, unfold map_rec_term, parametricity)

lemma param_map_induct1:
  assumes "(xs,xs') \<in> \<langle>R1\<rangle>list_rel"
      and "\<And>x y. x \<in> set xs \<Longrightarrow> (x,y) \<in> R1 \<Longrightarrow> (f x, f' y) \<in> R2"
    shows "(map f xs, map f' xs')\<in> \<langle>R2\<rangle>list_rel"
using assms by (induction xs xs' rule: list_rel_induct) auto

lemma subst_apply_term_autoref[autoref_rules, param]:
  "(subst_apply_term, subst_apply_term) \<in> \<langle>F,V\<rangle>term_rel \<rightarrow> (V \<rightarrow> \<langle>F,W\<rangle>term_rel) \<rightarrow> \<langle>F,W\<rangle>term_rel"
proof (intro fun_relI, goal_cases)
  note [param] = param_map_induct1
  case (1 t t')
    thus ?case by (induction t arbitrary: t'; elim term_relE, simp, parametricity)
qed

consts i_term :: "interface \<Rightarrow> interface \<Rightarrow> interface"
lemmas [autoref_rel_intf] = REL_INTFI[of term_rel i_term]

lemma [autoref_itype]:
  "Var ::\<^sub>i V \<rightarrow>\<^sub>i \<langle>F,V\<rangle>\<^sub>ii_term"
  "Fun ::\<^sub>i F \<rightarrow>\<^sub>i \<langle>\<langle>F,V\<rangle>\<^sub>ii_term\<rangle>\<^sub>ii_list \<rightarrow>\<^sub>i \<langle>F,V\<rangle>\<^sub>ii_term"
  "case_term ::\<^sub>i  (V \<rightarrow>\<^sub>i X) \<rightarrow>\<^sub>i (F \<rightarrow>\<^sub>i \<langle>\<langle>F,V\<rangle>\<^sub>ii_term\<rangle>\<^sub>ii_list \<rightarrow>\<^sub>i X) \<rightarrow>\<^sub>i \<langle>F,V\<rangle>\<^sub>ii_term \<rightarrow>\<^sub>i X"
  "rec_term ::\<^sub>i (V \<rightarrow>\<^sub>i X) \<rightarrow>\<^sub>i (F \<rightarrow>\<^sub>i \<langle>\<langle>\<langle>F,V\<rangle>\<^sub>ii_term, X\<rangle>\<^sub>ii_prod\<rangle>\<^sub>ii_list \<rightarrow>\<^sub>i X) \<rightarrow>\<^sub>i \<langle>F,V\<rangle>\<^sub>ii_term \<rightarrow>\<^sub>i X"
  "map_term ::\<^sub>i (F \<rightarrow>\<^sub>i F') \<rightarrow>\<^sub>i (V \<rightarrow>\<^sub>i V') \<rightarrow>\<^sub>i \<langle>F,V\<rangle>\<^sub>ii_term \<rightarrow>\<^sub>i \<langle>F',V'\<rangle>\<^sub>ii_term"
  "subst_apply_term ::\<^sub>i \<langle>F,V\<rangle>\<^sub>ii_term \<rightarrow>\<^sub>i (V \<rightarrow>\<^sub>i  \<langle>F,W\<rangle>\<^sub>ii_term) \<rightarrow>\<^sub>i \<langle>F,W\<rangle>\<^sub>ii_term"
by simp_all

subsection {* Relator for @{typ "('f,'v)ctxt"} *}

inductive_set ctxt_rel_aux for Rs Rv where
  Hole[intro,simp]: "(\<box>, \<box>) \<in> ctxt_rel_aux Rs Rv"
| More:"\<lbrakk> (f,f') \<in> Rs; 
         (ss1, ss1') \<in> \<langle>\<langle>Rs,Rv\<rangle>term_rel\<rangle>list_rel;
         (C,C') \<in> ctxt_rel_aux Rs Rv;
         (ss2, ss2') \<in> \<langle>\<langle>Rs,Rv\<rangle>term_rel\<rangle>list_rel\<rbrakk>
         \<Longrightarrow> (More f ss1 C ss2, More f' ss1' C' ss2') \<in> ctxt_rel_aux Rs Rv"

definition ctxt_rel_def_internal: "ctxt_rel \<equiv> ctxt_rel_aux"
lemma ctxt_rel_def[refine_rel_defs]: 
  "\<langle>Rs, Rv\<rangle>ctxt_rel \<equiv> ctxt_rel_aux Rs Rv"
by (simp add: ctxt_rel_def_internal relAPP_def)

lemma ctxt_relE1:
  "(\<box>, C) \<in> \<langle>F,V\<rangle>ctxt_rel \<Longrightarrow> (C = \<box> \<Longrightarrow> P) \<Longrightarrow> P"
by (auto simp: ctxt_rel_def elim: ctxt_rel_aux.cases)

lemma ctxt_relE2:
  "(C, \<box>) \<in> \<langle>F,V\<rangle>ctxt_rel \<Longrightarrow> (C = \<box> \<Longrightarrow> P) \<Longrightarrow> P"
by (auto simp: ctxt_rel_def elim: ctxt_rel_aux.cases)

lemma ctxt_relE3:
  assumes "(More f ss1 c ss2, C) \<in> \<langle>F,V\<rangle>ctxt_rel"
  obtains f' ss1' c' ss2' where "C = More f' ss1' c' ss2'" "(f,f') \<in> F" "(c,c') \<in> \<langle>F,V\<rangle>ctxt_rel"
    "(ss1, ss1') \<in> \<langle>\<langle>F,V\<rangle>term_rel\<rangle>list_rel" "(ss2, ss2') \<in> \<langle>\<langle>F,V\<rangle>term_rel\<rangle>list_rel"
using assms by (auto simp: ctxt_rel_def elim: ctxt_rel_aux.cases)

lemma ctxt_relE4:
  assumes "(C, More f ss1 c ss2) \<in> \<langle>F,V\<rangle>ctxt_rel"
  obtains f' ss1' c' ss2' where "C = More f' ss1' c' ss2'" "(f',f) \<in> F" "(c',c) \<in> \<langle>F,V\<rangle>ctxt_rel"
    "(ss1', ss1) \<in> \<langle>\<langle>F,V\<rangle>term_rel\<rangle>list_rel" "(ss2', ss2) \<in> \<langle>\<langle>F,V\<rangle>term_rel\<rangle>list_rel"
using assms by (auto simp: ctxt_rel_def elim: ctxt_rel_aux.cases)

lemmas ctxt_relE = ctxt_relE1 ctxt_relE2 ctxt_relE3 ctxt_relE4

lemma ctxt_rel_sv[relator_props]:
  assumes "single_valued Rs" "single_valued Rv"
  shows "single_valued (\<langle>Rs, Rv\<rangle>ctxt_rel)"
unfolding ctxt_rel_def proof (intro single_valuedI allI impI)
  fix x y z
  assume "(x, y) \<in> ctxt_rel_aux Rs Rv" and "(x, z) \<in> ctxt_rel_aux Rs Rv"
  thus "y = z"
  proof (induction arbitrary: z)
    case More
      note sv_intros = single_valuedD list_rel_sv term_rel_sv More.IH More.hyps
      from More.prems assms show ?case by (auto elim!: ctxt_rel_aux.cases[of _ z] intro: sv_intros)
  qed (auto elim: ctxt_rel_aux.cases)
qed

lemma ctxt_rel_id[relator_props]:
  assumes "Rs = Id" and "Rv = Id"
    shows "\<langle>Rs, Rv\<rangle>ctxt_rel = Id"
unfolding ctxt_rel_def proof(simp only: assms, rule)
  show "ctxt_rel_aux Id Id \<subseteq> Id"
  proof
    fix x y :: "('f, 'v) ctxt"
    assume "(x,y) \<in> ctxt_rel_aux Id Id"
    hence "x = y" by (induction, simp_all add: list_rel_id)
    thus "(x,y) \<in> Id" by blast
  qed
  show "Id \<subseteq> ctxt_rel_aux Id Id"
  proof
    fix x y :: "('f, 'v) ctxt"
    assume "(x,y) \<in> Id"
    moreover have "(x,x) \<in> ctxt_rel_aux Id Id" by (induction x, auto intro: More)
    ultimately show "(x,y) \<in> ctxt_rel_aux Id Id" by simp
  qed
qed

lemma ctxt_rel_id_simp[simp]:
  shows "\<langle>Id, Id\<rangle>ctxt_rel = Id" by tagged_solver

lemma ctxt_rel_mono[relator_props]:
  assumes "F \<subseteq> F'" and "V \<subseteq> V'"
  shows "\<langle>F, V\<rangle>ctxt_rel \<subseteq> \<langle>F', V'\<rangle>ctxt_rel"
unfolding ctxt_rel_def proof
  fix x y
  assume "(x, y) \<in> ctxt_rel_aux F V"
  thus "(x, y) \<in> ctxt_rel_aux F' V'" proof (induction)
    case More
      with assms show ?case by (intro ctxt_rel_aux.More, auto dest: list_rel_mono[OF term_rel_mono])
  qed simp
qed

lemma ctxt_rel_range[relator_props]:
  assumes sym: "Range F = UNIV" and "Range V = UNIV"
    shows "Range (\<langle>F, V\<rangle>ctxt_rel) = UNIV"
unfolding ctxt_rel_def proof
  show "Range (ctxt_rel_aux F V) \<subseteq> UNIV" by blast
  show "UNIV \<subseteq> Range (ctxt_rel_aux F V)"
  proof
    fix x
    show "x \<in> Range (ctxt_rel_aux F V)" unfolding Range_iff proof (induction x)
      case (More f ss1 C ss2)
        note range = list_rel_range[OF term_rel_range, OF assms]
        from sym obtain f' where "(f',f) \<in> F" by blast
        moreover from range obtain ss1' ss2' where
          "(ss1',ss1) \<in> \<langle>\<langle>F, V\<rangle>term_rel\<rangle>list_rel" "(ss2',ss2) \<in> \<langle>\<langle>F, V\<rangle>term_rel\<rangle>list_rel" by blast
        moreover from More.IH obtain C' where "(C', C) \<in> ctxt_rel_aux F V" by blast
        ultimately show ?case by (intro exI[of _ "More f' ss1' C' ss2'"], auto intro: ctxt_rel_aux.More)
    qed (intro exI[of _ \<box>], simp)
  qed
qed

lemma ctxt_rel_autoref[autoref_rules, param]:
  "(\<box>,\<box>) \<in> \<langle>F,V\<rangle>ctxt_rel"
  "(More,More) \<in> F \<rightarrow> \<langle>\<langle>F,V\<rangle>term_rel\<rangle>list_rel \<rightarrow> \<langle>F,V\<rangle>ctxt_rel \<rightarrow> \<langle>\<langle>F,V\<rangle>term_rel\<rangle>list_rel\<rightarrow> \<langle>F,V\<rangle>ctxt_rel"
  "(case_ctxt,case_ctxt) \<in> X \<rightarrow> (F \<rightarrow> \<langle>\<langle>F,V\<rangle>term_rel\<rangle>list_rel \<rightarrow> \<langle>F,V\<rangle>ctxt_rel \<rightarrow> \<langle>\<langle>F,V\<rangle>term_rel\<rangle>list_rel \<rightarrow> X) \<rightarrow> \<langle>F,V\<rangle>ctxt_rel \<rightarrow> X"
unfolding ctxt_rel_def
by (force intro: More split: ctxt.split elim: ctxt_rel_aux.cases dest: fun_relD)+

lemma ctxt_compose_autoref[autoref_rules, param]:
  "(op \<circ>\<^sub>c, op \<circ>\<^sub>c) \<in> \<langle>F,V\<rangle>ctxt_rel \<rightarrow> \<langle>F,V\<rangle>ctxt_rel \<rightarrow> \<langle>F,V\<rangle>ctxt_rel"
proof (intro fun_relI, goal_cases)
  case (1 a a')
  thus ?case by (induction a arbitrary: a' rule: ctxt.induct; elim ctxt_relE, simp, parametricity?)
qed

lemma ctxt_apply_term_autoref[autoref_rules, param]:
  "(ctxt_apply_term, ctxt_apply_term) \<in> \<langle>F,V\<rangle>ctxt_rel \<rightarrow> \<langle>F,V\<rangle>term_rel \<rightarrow> \<langle>F,V\<rangle>term_rel"
proof (intro fun_relI, goal_cases)
  case (1 a a')
  thus ?case by (induction a arbitrary: a' rule: ctxt.induct; elim ctxt_relE, simp, parametricity?)
qed

end
